@@ -8,9 +8,9 @@ package Text::Layout;
 
 use Carp;
 
+use Text::Layout::Version;
 
-
-our $VERSION = "0.018.3";
+our $VERSION = $Text::Layout::Version::VERSION;
 
 =head1 NAME
 
@@ -48,6 +48,9 @@ Example, using PDF::API2 integration:
     # Create a markup instance.
     my $layout = Text::Layout->new($pdf);
 
+    # This example uses PDF corefonts only.
+    Text::Layout::FontConfig->register_corefonts;
+
     $layout->set_font_description(Text::Layout::FontConfig->from_string("times 40"));
     $layout->set_markup( q{The <i><span foreground="red">quick</span> <span size="20"><b>brown</b></span></i> fox} );
 
@@ -88,32 +91,201 @@ PDF::API2 uses the coordinate system as defined in the PDF
 specification. It starts off bottom left. For western text the
 direction is increasing I<x> and B<de>creasing I<y>.
 
+=head1 Pango Conformance Mode
+
+Text::Layout can operate in one of two modes: I<convenience mode>
+(enabled by default), and I<Pango conformance mode>. The desired mode
+can be selected by calling the method set_pango_scaling().
+
 =head2 Pango coordinates
 
 Pango uses two device coordinates units: Pango units and device units.
-Pango units are 1000 (C<PANGO_SCALE>) times the device units.
+Pango units are 1024 (C<PANGO_SCALE>) times the device units.
 
 Several methods have two variants, e.g. get_size() and
 get_pixel_size(). The pixel-variant uses device units while the other
 variant uses Pango units.
 
-This module assumes no scaling. If you insist on multiplying all
-values by PANGO_SCALE use the set_pango_scale() method to say so and
-we'll divide everything back again internally.
+In I<convenience mode>, this module assumes no scaling. All units are
+PDF device units (1/72 inch).
 
 =head2 Pango device units
 
-Pango device units are 96dpi while PDF uses 72dpi. This module ignores
-this and uses PDF units everywhere, except for font sizes. Since we
-want e.g. a C<Times 20> font to be of equal size in the two systems,
-it will set the PDF font size to 15.
+Device units are used for font rendering. Pango device units are 96dpi
+while PDF uses 72dpi.
 
-I<DBD: Is this really a good idea?>
+In I<convenience mode> this is ignored. E.g. a C<Times 20> font 
+will be of equal size in the two systems,
+
+In I<Pango conformance mode> you would need to specify a font size of
+C<15360> to get a 20pt font.
+
+=head1 SUPPORTED MARKUP
+
+Text::Layout recognizes most of the Pango markup as provided by the
+Pango library version 1.50 or newer. However, not everything is supported.
+
+=head2 Span attributes
+
+=over 8
+
+=item font="I<DESC>"   font_desc="I<DESC>"
+
+Specifies a font to be used, e.g. C<Serif 20>.
+
+=item font_face="I<FAM>"   face="I<FAM>"
+
+Specifies a font family to be used.
+
+=item font_family="I<FAM>"
+
+Same as font_face="I<FAM>".
+
+=item size=I<FNUM>   size=I<FNUM>pt   size=I<FNUM>%
+
+Font size in 1024ths of a point (conformance mode), or in points (e.g.
+‘12.5pt’), or a percentage (e.g. ‘200%’), or one of the relative sizes
+‘smaller’ or ‘larger’.
+
+Note that in Pango conformance mode, the actual font size is 96/72
+larger. So C<"45pt"> gives a 60pt font.
+
+=item style="I<STYLE>"   font_style="I<STYLE>"
+
+Specifes the font style, e.g. C<italic>.
+
+=item weight="I<WEIGHT>"   font_weight="I<WEIGHT>"
+
+Specifies the font weight, e.g. C<bold>.
+
+=item foreground="I<COLOR>"   fgcolor="I<COLOR>"   color="I<COLOR>"
+
+Specifies the foreground colour, e.g. C<black>.
+
+=item background="I<COLOR>"   bgcolor="I<COLOR>"
+
+Specifies the background colour, e.g. C<white>.
+
+=item underline="I<TYPE>"
+
+Enables underlining.
+I<TYPE> must be C<none>, C<single>, or C<double>.
+
+=item underline_color="I<COLOR>"
+
+The colour to be used for underlining, if enabled.
+
+=item overline="I<TYPE>"
+
+Enables overlining.
+I<TYPE> must be C<none>, C<single>, or C<double>.
+
+=item overline_color="I<COLOR>"
+
+The colour to be used for ovderlining, if enabled.
+
+=item strikethrough="I<ARG>"
+
+Enables or diables overlining. I<ARG> must be C<true> or C<1> to
+enable, and C<false> or C<0> to disable.
+
+=item strikethrough_color="I<COLOR>"
+
+The colour to be used for striking, if enabled.
+
+=item rise=C<NUM>
+
+In convenience mode, lowers the text by I<NUM>/1024 of the font size.
+May be negative to rise the text.
+
+In Pango conformance mode, rises the text by I<NUM> units from the baseline.
+May be negative to lower the text.
+
+Note: In Pango conformance mode, C<rise> does B<not> accumulate.
+ Use C<baseline_shift> instead.
+
+=item rise=C<NUM>pt   rise=C<NUM>%   rise=C<NUM>em   rise=C<NUM>ex
+
+Rises the text from the baseline. May be negative to lower the text.
+
+Units are points if postfixed by B<pt>, and a percentage of the current font size if postfixed by B<%>.
+
+B<em> units are equal to the current font size, B<ex> half the font size.
+
+Note: This is not (yet?) part of the Pango markup standard.
+
+=item baseline_shift=C<NUM>   beseline_shift=C<NUM>pt   baseline_shift=C<NUM>%
+
+Like C<rise>, but accumulates.
+
+=back
+
+Also supported but not part of the official Pango Markup specification.
+
+=over 8
+
+=item href="C<URL>"
+
+Creates a clickable target that activates the I<URL>.
+
+=back
+
+=head2 Shortcuts
+
+Equivalent C<span> attributes for shortcuts.
+
+=over 8
+
+=item b
+
+weight=bold
+
+=item big
+
+larger
+
+=item emp
+
+style=italic
+
+=item i
+
+style=italic
+
+=item s
+
+strikethrough=true
+
+=item small
+
+size=smaller
+
+=item strong
+
+weight=bold
+
+=item sub
+
+size=smaller rise=-30%
+
+=item sup
+
+size=smaller rise=30%
+
+=item tt
+
+face=monospace
+
+=item u
+
+underline=single
+
+=back
 
 =cut
 
 use constant {
-    PANGO_SCALE		=> 1000,
+    PANGO_SCALE		=> 1024,
     PANGO_DEVICE_UNITS	=>   96,
     PDF_DEVICE_UNITS	=>   72,
 };
@@ -130,9 +302,16 @@ use Text::Layout::FontConfig;
 
 =item new( $pdf )
 
-Creates a new layout instance.
+Creates a new layout instance for PDF::API2. This is for convenience.
+It is equivalent to
 
-The arguments are passed to the backend. Read its documentation.
+    use Text::Layout::PDFAPI2;
+    $layout = Text::Layout::PDFAPI2->new($pdf);
+
+For other implementations only the above method can be used.
+
+The argument is the I<context> for text formatting. In the case of
+PDF::API2 this will be a PDF::API2 object.
 
 =back
 
@@ -141,30 +320,22 @@ The arguments are passed to the backend. Read its documentation.
 sub new {
     my ( $pkg, @data ) = @_;
 
-    my $be = __PACKAGE__."::Markdown";
-
-    if ( @data >= 2 && $data[0] eq "backend") {
-	shift;
-	$be = shift;
+    $pkg =~ s/::$//;
+    if ( $pkg eq __PACKAGE__ ) {
+	# For convenience (and backward compatibility)...
+	unless ( @data >= 1 && ref($data[0]) =~ /^PDF::(API2|Builder)\b/ ) {
+	    croak("Please instanciate a backend, e.g. ".__PACKAGE__."::PDFAPI2");
+	}
+	require Text::Layout::PDFAPI2;
+	return Text::Layout::PDFAPI2->new(@data); # CUL8R
     }
-    if ( $INC{"PDF/API2.pm"} || $INC{"PDF/Builder.pm"} ) {
-	$be = __PACKAGE__."::PDFAPI2";
-    }
-    elsif ( $INC{"Cairo.pm"} ) {
-	$be = __PACKAGE__."::Cairo";
-    }
-    elsif ( $INC{"Pango.pm"} ) {
-	$be = __PACKAGE__."::Pango";
-    }
-    eval "require $be" or croak($@);
-    Text::Layout::FontConfig->set_loader( $be->can("load_font") );
-    bless { _be      => $be,
-	    _context => $be->init(@data),
+    bless { _context => undef,
 	    _fonts   => {},
 	    _content => [],
 	    _px2pu   => \&px2pu,
 	    _pu2px   => \&pu2px,
-	  };
+	    _pango   => 0,
+	  } => $pkg;
 }
 
 =over
@@ -198,7 +369,7 @@ sub copy {
 
 =item get_context
 
-Gets the context (PDF document) of this layout.
+Gets the context of this layout.
 
 =back
 
@@ -250,12 +421,14 @@ layout from the markup as this function does not clear all attributes.
 sub set_text {
     my ( $self, $string ) = @_;
     $self->{_content} =
-      [ { text  => $string,
+      [ { type  => "text",
+	  text  => $string,
 	  font  => $self->{_currentfont},
 	  size  => $self->{_currentsize},
 	  color => $self->{_currentcolor},
 	  base  => 0,
 	} ];
+    delete( $self->{_bbcache} );
 }
 
 =over
@@ -314,26 +487,37 @@ Implementation note: Although all markup is parsed, not all is implemented.
 
 =cut
 
-use constant STEP => 0.8;	# TODO: Optimal value?
+use Text::ParseWords;
+use constant STEP => 0.83333125;	# Cf. Pango
 
 my %magstep =
   ( "xx-small"	=> 1.0*STEP*STEP*STEP,
     "x-small"	=> 1.0*STEP*STEP,
     "small"	=> 1.0*STEP,
+    "smaller"	=> 1.0*STEP,
     "medium"	=> 1.0,
     "large"	=> 1.0/STEP,
+    "larger"	=> 1.0/STEP,
     "x-large"	=> 1.0/(STEP*STEP),
     "xx-large"  => 1.0/(STEP*STEP*STEP),
   );
 
 sub set_markup {
     my ( $self, $string ) = @_;
-
+    confess("set_markup with UNDEF") unless defined $string;
     my @stack;
     my @content;
     my $fcur = $self->{_currentfont};
     my $fcol = $self->{_currentcolor};
     my $fsiz = $self->{_currentsize};
+    my $href;
+    my $undl;
+    my $uncl = $fcol;
+    my $ovrl;
+    my $ovcl = $fcol;
+    my $strk;
+    my $stcl = $fcol;
+    my $bcol;
     my $base = 0;
 
     my $try_size = sub {
@@ -356,48 +540,68 @@ sub set_markup {
     my $span;
     $span = sub {
 	my ( $v ) = @_;
-	# Split the span attributes.
-	foreach my $k ( split(' ',$v) ) {
+	# Split the span attributes. Note that shellwords is so kind to
+	# split 'font="xx yy"' as a single word 'font=xx yy'.
+	foreach my $k ( shellwords($v) ) {
 
 	    # key=value
 	    if ( $k =~ /^([-\w]+)=(.+)$/ ) {
 		my ( $k, $v ) = ( $1, $2 );
 
-		# Strip quotes.
-		$v =~ s/^(["'])(.*)\1$/$2/;
+		# Ignore case unless required.
+		$v = lc $v unless $k =~ /^(link|href|a)$/;
+
+		# Strip quotes. Shouldn't be necessary now we use shellwords.
+		# $v =~ s/^(["'])(.*)\1$/$2/;
 
 		# <span font_desc="Sans 20">
-		if ( $k =~ /^(font_desc)$/ ) {
-		    # NYI.
+		if ( $k =~ /^(font|font_desc)$/ ) {
+		    $fcur = Text::Layout::FontConfig->from_string($v);
+		    $fsiz = $fcur->get_size if $fcur->get_size;
 		}
 
 		# <span face="Sans">
-		elsif ( $k =~ /^(face|font_family)$/ ) {
-		    $fcur = Text::Layout::FontConfig->find_font($v);
+		elsif ( $k =~ /^(face|font_face|font_family)$/ ) {
+		    $fcur = Text::Layout::FontConfig->find_font( $v,
+								 $fcur->{style},
+								 $fcur->{weight});
 		}
 
 		# <span size=20>
-		elsif ( $k eq "size" ) {
+		elsif ( $k =~ /^(size|font_size)$/ ) {
 		    if ( $try_size->($v) ) {
 			#ok
 		    }
-		    elsif ( $v =~ /\d+(?:\.\d+)?$/ ) {
-			$fsiz = 0 + $v;
+		    elsif ( $v =~ /^(\d+(?:\.\d+)?)pt$/ ) {
+			$fsiz = $self->{_pango}
+			  ? $1 * PANGO_DEVICE_UNITS / PDF_DEVICE_UNITS
+			  : $1;
+			# warn("fsiz \"$v\" -> $fsiz\n");
+		    }
+		    elsif ( $v =~ /^\d+(?:\.\d+)?$/ ) {
+			$fsiz = $self->{_pango}
+			  ? $v * PANGO_DEVICE_UNITS / PDF_DEVICE_UNITS / PANGO_SCALE
+			  : $v;
+			# warn("fsiz \"$v\" -> $fsiz\n");
+		    }
+		    elsif ( $v =~ /^(\d+(?:\.\d+)?)\%$/ ) {
+			$fsiz *= $1 / 100;
+			# warn("fsiz \"$v\" -> $fsiz\n");
 		    }
 		    else {
-			carp("Invalid size: $v\n");
+			carp("Invalid size: \"$v\"\n");
 		    }
 		}
 
 		# <span style="Italic">
-		elsif ( $k eq "style" ) {
+		elsif ( $k =~ /^(style|font_style)$/ ) {
 		    $v = Text::Layout::FontConfig::_norm_style($v);
 		    $fcur = Text::Layout::FontConfig->find_font( $fcur->{family},
 					      $v, $fcur->{weight} );
 		}
 
 		# <span weight="bold">
-		elsif ( $k eq "weight" ) {
+		elsif ( $k =~ /^(weight|font_weight)$/ ) {
 		    $v = Text::Layout::FontConfig::_norm_weight($v);
 		    $fcur = Text::Layout::FontConfig->find_font( $fcur->{family},
 					      $fcur->{style}, $v );
@@ -405,34 +609,82 @@ sub set_markup {
 
 		# <span variant="...">
 		# <span stretch="...">
-		elsif ( $k =~ /^variant|stretch$/ ) {
+		elsif ( $k =~ /^(?:font_)?(variant|stretch)$/ ) {
+		    # ignore for now
+		}
+		elsif ( $k =~ /^(features|background_alpha|alpha)$/ ) {
 		    # ignore for now
 		}
 
 		# <span foreground="red">
-		# We allow 'color' as an alternative.
-		elsif ( $k =~ /^foreground|color$/ ) {
+		elsif ( $k =~ /^(foreground|fgcolor|color)$/ ) {
 		    $fcol = $v;
 		}
 
 		# <span background="red">
-		elsif ( $k eq "background" ) {
-		    # NYI.
+		elsif ( $k =~ /^(background|bgcolor)$/ ) {
+		    $bcol = $v;
 		}
 
 		# <span underline="double">
-		elsif ( $k eq "underline" ) {
-		    # NYI.
+		elsif ( $k eq "underline" && $v =~ /^(none|single|double)$/i ) {
+		    $undl = lc $v;
+		}
+		elsif ( $k eq "underline_color" ) {
+		    $uncl = lc $v;
+		}
+
+		# <span overline="double">
+		elsif ( $k eq "overline" && $v =~ /^(none|single|double)$/i ) {
+		    $ovrl = lc $v;
+		}
+		elsif ( $k eq "overline_color" ) {
+		    $ovcl = lc $v;
 		}
 
 		# <span rise=324>
 		elsif ( $k eq "rise" ) {
-		    $base += $v / 1024 * $fsiz;
+		    if ( $v =~ /^(-?\d+(?:\.\d*)?)pt$/ ) {
+			$base = -$1;
+		    }
+		    elsif ( !$self->{_pango} && $v =~ /^(-?\d+(?:\.\d*)?)\%$/ ) {
+			$base = -$1 * $fsiz / 100;
+		    }
+		    elsif ( !$self->{_pango} && $v =~ /^(-?\d+(?:\.\d*)?)e([mx])$/ ) {
+			$base = $2 eq 'x' ? -$1 * $fsiz / 2 : -$1 * $fsiz;
+		    }
+		    else {
+			$v /= PANGO_SCALE;
+			$base = $self->{_pango} ? -$v : $base + $v * $fsiz;
+		    }
+		}
+
+		# <span baseline_shift=324>
+		# Line rise, but accumulate.
+		elsif ( $k eq "baseline_shift" ) {
+		    if ( $v =~ /^(-?\d+(?:\.\d*)?)pt$/ ) {
+			$base -= $1;
+		    }
+		    elsif ( $v =~ /^(-?\d+(?:\.\d*)?)\%$/ ) {
+			$base -= $1 * $fsiz / 100;
+		    }
+		    elsif ( $v =~ /^(-?\d+(?:\.\d*)?)e([mx])$/ ) {
+			$base = $2 eq 'x' ? $1 * $fsiz / 2 : $1 * $fsiz;
+		    }
+		    else {
+			$base -= $self->{_pango} ? $v : -$v * $fsiz;
+		    }
 		}
 
 		# <span strikethrough=false>
-		elsif ( $k eq "strikethrough" ) {
-		    # NYI.
+		elsif ( $k eq "strikethrough" && $v =~ /^(true|1)$/i ) {
+		    $strk = 1;
+		}
+		elsif ( $k eq "strikethrough" && $v =~ /^(false|0)$/i ) {
+		    $strk = 0;
+		}
+		elsif ( $k eq "strikethrough_color" ) {
+		    $stcl = $v;
 		}
 
 		# <span fallback=false>
@@ -444,6 +696,11 @@ sub set_markup {
 		elsif ( $k eq "lang" ) {
 		    # Not supported.
 		}
+
+		# <span href="...">
+		elsif ( $k eq "href" ) {
+		    $href = $v;
+		}
 	    }
 
 	    # <span strikethrough>
@@ -457,13 +714,57 @@ sub set_markup {
 	    }
 
 	    else {
-		carp("Invalid span markup: $k\n");
+		carp("Invalid span markup: \"$k\"\n");
 	    }
 	}
     };
 
+    my $image = sub {
+	my ( $v ) = @_;
+
+	my %img;
+	# Split the attributes.
+	foreach my $k ( shellwords($v) ) {
+
+	    # key=value
+	    if ( $k =~ /^([-\w]+)=(.+)$/ ) {
+		my ( $k, $v ) = ( $1, $2 );
+
+		# Ignore case unless required.
+		$v = lc $v unless $k =~ /^(src|id)$/;
+
+		# Strip quotes. Shouldn't be necessary now we use shellwords.
+		# $v =~ s/^(["'])(.*)\1$/$2/;
+
+		if ( $k eq "src" ) {
+		    $img{src} = $v;
+		}
+		elsif ( $k eq "id" ) {
+		    $img{id} = $v;
+		}
+		elsif ( $k =~ /^(width|height|w|h)$/ ) {
+		    $img{$k} = $v;
+		}
+		elsif ( $k =~ /^(x|y)$/ ) {
+		    $img{$k} = $v;
+		}
+		elsif ( $k eq "border" ) {
+		    $img{border} = $v;
+		}
+		else {
+		    carp("Invalid image attribute: \"$k\"\n");
+		}
+	    }
+	    else {
+		carp("Invalid image attribute: \"$k\"\n");
+	    }
+	}
+	return \%img;
+    };
+
+
     # Split the string on markup instructions.
-    foreach my $a ( split( /(<.*?>)/, $string ) ) {
+ L: foreach my $a ( split( /(<.*?>)/, $string ) ) {
 
 	# Closing markup, e.g. </b> or </span>.
 	if ( $a =~ m;^<\s*/\s*(\w+)(.*)>$; ) {
@@ -472,10 +773,9 @@ sub set_markup {
 		# Check if it is closing the currently pending markup.
 		if ( $stack[-1][0] =~ /^<\s*$k\b/ ) {
 		    # Restore.
-		    $fcur = $stack[-1][1];
-		    $fsiz = $stack[-1][2];
-		    $fcol = $stack[-1][3];
-		    $base = $stack[-1][4];
+		    ( undef,
+		      $fcur, $fsiz, $fcol, $undl, $uncl, $ovrl, $ovcl,
+		      $bcol, $strk, $stcl, $base, $href ) = @{$stack[-1]};
 		    pop(@stack);
 		}
 		else {
@@ -492,11 +792,15 @@ sub set_markup {
 	}
 
 	# Opening markup, e.g. <b> or <span ...>.
-	elsif ( $a =~ m;^<\s*([-\w]+)(.*)>$; ) {
+	elsif ( $a =~ m;^<\s*([-\w]+)(.*)?>$; ) {
 	    my $k = lc $1;
-	    my $v = lc $2;
+	    my $v = $2;
+	    my $closed = $v =~ s;\s*/\s*$;;;
+
 	    # Save.
-	    push( @stack, [ "<$k$v>", $fcur, $fsiz, $fcol, $base ] );
+	    push( @stack, [ "<$k".lc($v).">",
+			    $fcur, $fsiz, $fcol, $undl, $uncl, $ovrl, $ovcl,
+			    $bcol, $strk, $stcl, $base, $href ] );
 
 	    # <b> <strong>
 	    if ( $k =~ /^(b|strong)$/ ) {
@@ -520,12 +824,12 @@ sub set_markup {
 
 	    # <sub>
 	    elsif ( $k eq "sub" ) {
-		$span->("size=smaller rise=300");
+		$span->("size=smaller rise=-30%");
 	    }
 
 	    # <sup>
 	    elsif ( $k eq "sup" ) {
-		$span->("size=smaller rise=-300");
+		$span->("size=smaller rise=30%");
 	    }
 
 	    # <small>
@@ -548,20 +852,53 @@ sub set_markup {
 		$span->($v);
 	    }
 
+	    # <.../>.
+	    elsif ( ( my $p = $self->get_element_handler($k) ) && $closed ) {
+		push( @content, { type => $k,
+				  %{ $p->parse
+				       ( { type			      => $k,
+					   font			      => $fcur,
+					   size			      => $fsiz,
+					   color		      => $fcol,
+					   bgcolor		      => $bcol,
+					   underline		      => $undl,
+					   underline_color	      => $uncl,
+					   overline		      => $ovrl,
+					   overline_color	      => $ovcl,
+					   strikethrough	      => $strk,
+					   strikethrough_color	      => $stcl,
+					   base			      => $base,
+					   href			      => $href,
+					 }, $k, $v ) } } );
+	    }
+
 	    else {
-		carp("Invalid markup: $k\n");
+		carp("Invalid markup: \"$k\"\n");
+	    }
+	    if ( $closed ) {
+		$a = "</$k>";
+		redo L;
 	    }
 	}
 
 	# Text.
 	else {
 	    push( @content,
-		  { text  => $a,
-		    font  => $fcur,
-		    size  => $fsiz,
-		    color => $fcol,
-		    base  => $base,
-		  } );
+		  { type		     => "text",
+		    text		     => $a,
+		    font		     => $fcur,
+		    size		     => $fsiz,
+		    color		     => $fcol,
+		    bgcolor		     => $bcol,
+		    underline		     => $undl,
+		    underline_color	     => $uncl,
+		    overline		     => $ovrl,
+		    overline_color	     => $ovcl,
+		    strikethrough	     => $strk,
+		    strikethrough_color	     => $stcl,
+		    base		     => $base,
+		    href		     => $href,
+		  } ) if defined $a && $a ne '';
 	}
     }
 
@@ -574,6 +911,7 @@ sub set_markup {
 
     # Store content.
     $self->{_content} = \@content;
+    delete( $self->{_bbcache} );
 }
 
 =over
@@ -625,9 +963,23 @@ sub set_font_description {
       unless UNIVERSAL::isa( $description, $o );
 
     $self->{_currentfont}  = $description;
-    $self->{_currentsize}  = $description->{size}
-      if $description->{size};
+
+    if ( my $sz = $description->{size} ) {
+	# Font sizes can be in either Pango units, or device units.
+	# Use heuristics.
+	if ( $sz > 2 * PANGO_SCALE ) { # Pango units
+	    $sz /= PANGO_SCALE;
+	}
+	if ( $self->{_pango} ) {
+	    # Pango uses 96dpi while PDF is 72dpi.
+	    $sz *= PANGO_DEVICE_UNITS / PDF_DEVICE_UNITS;
+	}
+	$self->{_currentsize} = $sz;
+    }
+
     $self->{_currentcolor} = $description->{color} || "black";
+
+    delete( $self->{_bbcache} );
 }
 
 =over
@@ -666,6 +1018,7 @@ Implementation note: Only alignment is implemented.
 sub set_width {
     my ( $self, $width ) = @_;
     $self->{_width} = $self->{_pu2px}->($width);
+    delete( $self->{_bbcache} );
 }
 
 =over
@@ -699,6 +1052,7 @@ Implementation note: Height restrictions are not yet implemented.
 sub set_height {
     my ( $self, $height ) = @_;
     $self->{_height} = $self->{_pu2px}->($height);
+    delete( $self->{_bbcache} );
 }
 
 =over
@@ -882,6 +1236,7 @@ Not yet implemented.
 sub set_spacing {
     my ( $self, $spacing ) = @_;
     $self->{_currentspacing} = $self->{_pu2px}->($spacing);
+    delete( $self->{_bbcache} );
 }
 
 =over
@@ -924,6 +1279,7 @@ Not yet implemented.
 sub set_line_spacing {
     my ( $self, $factor ) = @_;
     $self->{_currentlinespacing} = $factor;
+    delete( $self->{_bbcache} );
 }
 
 =over
@@ -1013,8 +1369,9 @@ sub set_alignment {
 	$self->{_alignment} = lc $align;
     }
     else {
-	croak("Invalid alignment: $align");
+	croak("Invalid alignment: \"$align\"");
     }
+    delete( $self->{_bbcache} );
 }
 
 =over
@@ -1146,19 +1503,22 @@ Computes the logical and ink extents of the layout.
 
 Logical extents are usually what you want for positioning things.
 
-Return value is an array (in list context) or an array ref (in scalar
-context) containing two hash refs with 4 values: C<x>, C<y>, C<width>,
-and C<height>. The first reflects the ink extents, the second the
-logical extents.
+Return value in scalar context is a hash ref with 4 values:
+ C<x>, C<y>, C<width>, and C<height>
+describing the logical extents of the layout.
+In list context an array of two hashrefs is returned.
+The first reflects the ink extents, the second the logical extents.
 
-C<x> will reflect the offset when text is centered or right aligned.
-It will be zero for left aligned text. For right aligned text, it will
-be the width of the layout.
+In the extents, C<x> will reflect the offset when text is centered or
+right aligned. It will be zero for left aligned text. For right
+aligned text, it will be the width of the layout.
 
 C<y> will reflect the offset when text is centered vertically or
 bottom aligned. It will be zero for top aligned text.
 
-Implementation note: Since the PDF support layer cannot calculate ink,
+See also get_pixel_extents below.
+
+Implementation note: If the PDF::API support layer cannot calculate ink,
 this function returns two identical extents.
 
 =back
@@ -1167,10 +1527,19 @@ this function returns two identical extents.
 
 sub get_extents {
     my ( $self ) = @_;
+    my $need_ink = wantarray;
 
-    my @bb = @{$self->get_bbox};
-    my $res = { x => $bb[0], y => 0, width => $bb[2], height => $bb[3]-$bb[1] };
-    return wantarray ? ( $res, $res ) : [ $res, $res ];
+    my @bb = $self->get_bbox($need_ink);
+    my $res = { bl    => $bb[0],
+		x     => $bb[1], y      => $bb[2],
+		width => $bb[3], height => $bb[4] };
+    my $ink = $res;
+    if ( @bb > 5 ) {
+	$ink = { bl    => $bb[0],
+		 x     => $bb[5], y      => $bb[6],
+		 width => $bb[7], height => $bb[8] };
+    }
+    return $need_ink ? ( $ink, $res ) : $res;
 }
 
 =over
@@ -1179,6 +1548,15 @@ sub get_extents {
 
 Same as get_extents, but using device units.
 
+The returned values are suitable for (assuming $pdf_text and
+$pdf_gfx are the PDF text and graphics contexts):
+
+    $layout->render( $x, $y, $pdf_text );
+    $box = $layout->get_pixel_extents;
+    $pdf_gfx->translate( $x, $y );
+    $pdf_gfx->rect( @$box{ qw( x y width height ) } );
+    $pdf_gfx->stroke;
+
 =back
 
 =cut
@@ -1186,9 +1564,19 @@ Same as get_extents, but using device units.
 sub get_pixel_extents {
     my ( $self ) = @_;
 
-    my @bb = @{$self->get_pixel_bbox};
-    my $res = { x => $bb[0], y => 0, width => $bb[2], height => $bb[3]-$bb[1] };
-    return wantarray ? ( $res, $res ) : [ $res, $res ];
+    my $need_ink = wantarray;
+
+    my @bb = $self->get_pixel_bbox($need_ink);
+    my $res = { bl    => $bb[0],
+		x     => $bb[1], y      => $bb[2],
+		width => $bb[3], height => $bb[4] };
+    my $ink = $res;
+    if ( @bb > 5 ) {
+	$ink = { bl    => $bb[0],
+		 x     => $bb[5], y      => $bb[6],
+		 width => $bb[7], height => $bb[8] };
+    }
+    return $need_ink ? ( $ink, $res ) : $res;
 }
 
 =over
@@ -1209,7 +1597,7 @@ returned.
 
 sub get_size {
     my ( $self ) = @_;
-    my $e = ($self->get_extents)[1];
+    my $e = $self->get_extents;
     wantarray
       ? return ( $e->{width}, $e->{height} )
       : return { width => $e->{width}, height => $e->{height} };
@@ -1227,7 +1615,7 @@ Same as get_size().
 
 sub get_pixel_size {
     my ( $self ) = @_;
-    my $e = ($self->get_pixel_extents)[1];
+    my $e = $self->get_pixel_extents;
     wantarray
       ? return ( $e->{width}, $e->{height} )
       : return { width => $e->{width}, height => $e->{height} };
@@ -1267,7 +1655,7 @@ See get_iter().
 
 sub get_baseline {
     my ( $self ) = @_;
-    return -$self->get_bbox->[3];
+    return -$self->get_bbox->[0];
 }
 
 =head1 METHODS NOT IMPLEMENTED
@@ -1319,6 +1707,7 @@ Sets the size for the current font.
 sub set_font_size {
     my ( $self, $size ) = @_;
     $self->{_currentsize} = $size;
+    delete( $self->{_bbcache} );
 }
 
 =over
@@ -1327,8 +1716,6 @@ sub set_font_size {
 
 Returns the size of the current font.
 
-NOTE: This is not a Pango API method.
-
 =back
 
 =cut
@@ -1342,50 +1729,52 @@ sub get_font_size {
 
 =item get_bbox
 
-Returns the bounding box of the text, w.r.t. the origin.
+Returns the bounding box of the text, w.r.t. the (top-left) origin.
 
-     +-----------+   ascend
-     |           |
-     |           |
-     |           |
-     |           |
-     |           |
-     | <-width-> |
-     |           |
-     |           |
-     o-----------+   o = origin baseline
-     |           |
-     |           |
-     +-----------+   descend
+bb = ( bl, x, y, width, height )
 
-bb = ( left, descend, right, ascend )
+bb[0] = baseline distance from the top.
 
-bb[0] = left is nonzero for centered and right aligned text
+bb[1] = displacement from the left, nonzero for centered and right aligned text
 
-bb[1] = descend is a negative value
+bb[2] = displacement from the top, usually zero
 
-bb[2] - bb[0] = advancewidth
+bb[3] = advancewidth
 
-bb[2] = layout width for right aligned text
+bb[4] = height
 
-bb[3] = ascend is a positive value
+Note that the bounding box will in general be equal to the font
+bounding box except for the advancewidth.
 
 NOTE: Some fonts do not include accents on capital letters in the ascend.
 
+If an argument is supplied and true, get_bbox() will attempt to
+calculate the ink extents as well, and add these as another set of 4
+elements,
+
+In list context returns the array of values, in scalar context an
+array ref.
+
 =back
 
 =cut
 
 sub get_pixel_bbox {
-    my ( $self ) = @_;
-    no strict 'refs';
-    my $res = "$self->{_be}::bbox"->( $self );
+    my ( $self, $all ) = @_;
+    my $res;
+    if ( $self->{_bbcache}
+	 && @{ $self->{_bbcache} } == ($all ? 9 : 5) ) {
+	$res = $self->{_bbcache};
+    }
+    else {
+	$res = $self->{_bbcache} = $self->bbox($all);
+    }
     wantarray ? @$res : $res;
 }
 
 sub get_bbox {
-    my ( $self ) = @_;
-    my @res = map { $self->{_px2pu}->($_) } ( $self->get_pixel_bbox );
+    my ( $self, $all ) = @_;
+    my @res = map { $self->{_px2pu}->($_) } ( $self->get_pixel_bbox($all) );
     wantarray ? @res : \@res;
 }
 
@@ -1398,7 +1787,7 @@ graphics environment.
 
 Use this instead of Pango::Cairo::show_layout().
 
-$text must be an object created by PDF $page->text method.
+For PDF::API2, $text must be an object created by the $page->text method.
 
 =back
 
@@ -1406,58 +1795,86 @@ $text must be an object created by PDF $page->text method.
 
 sub show {
     my ( $self, @data ) = @_;
-    no strict 'refs';
-    "$self->{_be}::render"->( $self, @data );
+    $self->render( @data );
 }
 
 =over
 
-=item set_pango_scale( $scale )
+=item set_pango_mode( $enable )
 
-Sets the pango scaling to $scale, which should be 1000.
+Enable/disable Pango conformance mode.
+See L<Pango Conformance Mode>.
 
-Set to 1 to cancel scaling, causing all methods to use pixel units
-instead of Pango units. Set to 0 to get the default behaviour.
+Returns the internal Pango scaling factor if enabled.
 
 =back
 
 =cut
 
-sub set_pango_scale {
-    my ( $self, $scale ) = @_;
-    $scale //= PANGO_SCALE;
-    if ( $scale == 0 ) {
-	$self->{_px2pu} = \&px2pu;
-	$self->{_pu2px} = \&pu2px;
-    }
-    elsif ( $scale == 1 ) {
-	$self->{_px2pu} = sub { $_[0] };
-	$self->{_pu2px} = sub { $_[0] };
-    }
-    else {
-	carp("Pango scale should be set to @{[PANGO_SCALE]} for Pango conformance")
-	  unless $scale == PANGO_SCALE;
-	$self->{_px2pu} = sub { $_[0] * $scale };
-	$self->{_pu2px} = sub { $_[0] / $scale };
+sub set_pango_mode {
+    my ( $self, $conformant ) = @_;
+
+    if ( $conformant ) {
+	delete( $self->{_bbcache} ) unless $self->{_pango};
+	$self->{_px2pu} = sub { $_[0] * PANGO_SCALE };
+	$self->{_pu2px} = sub { $_[0] / PANGO_SCALE };
+	return $self->{_pango} = PANGO_SCALE;
     }
+
+    delete( $self->{_bbcache} ) if $self->{_pango};
+    $self->{_px2pu} = \&px2pu;
+    $self->{_pu2px} = \&pu2px;
+    return $self->{_pango} = 0;
 }
 
+# Legacy.
+*set_pango_scale = \&set_pango_mode;
+
+=over
+
+=item get_pango
+
+See L<Pango Conformance Mode>.
+
+Returns the internal Pango scaling factor if conformance mode is
+enabled, otherwise it returns 1 (one).
+
+=back
+
+=cut
+
 sub get_pango_scale {
     my ( $self ) = @_;
-    $self->{_px2pu}->(1);
+    $self->{_pango} ? PANGO_SCALE : 1;
 }
 
 sub nyi {
     croak("Method \"" . (caller(1))[3] . "\" not implemented");
 }
 
+my $aliens;
+sub register_element {
+    my ( $self, $hd, @tags ) = @_;
+    croak("Element handler for \"$tags[0]\" does not play nicely")
+      unless $hd->DOES(qw(Text::Layout::ElementRole));
+    for ( @tags ) {
+	$aliens->{$_} = $hd;
+    }
+}
+
+sub get_element_handler {
+    my ( $self, $tag ) = @_;
+    return unless $tag;
+    $aliens->{$tag};
+}
+
 =head1 SEE ALSO
 
 Description of the Pango Markup Language:
-L<https://developer.gnome.org/pygtk/stable/pango-markup-language.html>.
+L<https://docs.gtk.org/Pango/pango_markup.html#pango-markup>.
 
 Documentation of the Pango Layout class:
-L<https://developer.gnome.org/pango/stable/pango-Layout-Objects.html>.
+L<https://docs.gtk.org/Pango/class.Layout.html>.
 
 L<PDF::API2>, L<PDF::Builder>, L<HarfBuzz::Shaper>, L<Font::TTF>.
 
