@@ -8,9 +8,9 @@ package Text::Layout::FontConfig;
 
 use Carp;
 
+use Text::Layout::Version;
 
-
-our $VERSION = "0.018.3";
+our $VERSION = $Text::Layout::VERSION;
 
 use Text::Layout::FontDescriptor;
 
@@ -41,6 +41,27 @@ my @dirs;
 my $loader;
 my $debug = 0;
 
+my $weights =
+  [ 100 => 'thin',
+    100 => 'hairline',
+    200 => 'extra light',
+    200 => 'ultra light',
+    300 => 'light',		# supported
+    350 => 'book',		# supported
+    400 => 'normal',		# supported
+    400 => 'regular',		# supported
+    500 => 'medium',		# supported
+    600 => 'semi bold',		# supported 'semi'
+    600 => 'demi bold',
+    700 => 'bold',		# supported
+    800 => 'extra bold',
+    800 => 'ultra bold',
+    900 => 'black',
+    900 => 'heavy',		# supported
+    950 => 'extra black',
+    950 => 'ultra black',
+  ];
+
 =head2 METHODS
 
 =over
@@ -66,16 +87,25 @@ If true, a predefined set of font names (the PDF corefonts) is registered.
 
 sub new {
     my ( $pkg, %atts ) = @_;
-    my $self = bless \my $i => $pkg;
+    my $self = bless {} => $pkg;
     if ( $atts{corefonts} ) {
 	$self->register_corefonts;
     }
     if ( $atts{loader} ) {
 	$loader = $atts{loader};
     }
+    $debug = $self->{debug} = $atts{debug};
     return $self;
 }
 
+sub reset {
+    my ( $self ) = @_;
+    %fonts = ();
+    @dirs = ();
+}
+
+sub debug { shift->{debug} }
+
 =over
 
 =item register_fonts( $font, $family, $style [ , $weight ] [ , $props ] )
@@ -110,8 +140,37 @@ Text shaping requires module L<HarfBuzz::Shaper>.
 
 =item *
 
-C<interline> - If set to a true value, an alternative way to determine
-glyph height is used. This may improve results for some fonts.
+C<ascender> - If set overrides the font ascender.
+This may be necessary to improve results for some fonts.
+The value is expressed in 1/1000th of an em.
+
+C<descender> - If set overrides the font descender.
+This may be necessary to improve results for some fonts.
+The value is expressed in 1/1000th of an em.
+
+=item *
+
+C<underline_thickness>, C<underline_position> - Overrides the font
+specified or calculated values for underline thickness and/or position.
+This may improve results for some fonts.
+
+=item *
+
+C<strikeline_thickness>, C<strikeline_position> - Overrides the font
+specified or calculated values for strikeline thickness and/or position.
+This may improve results for some fonts.
+
+Note that strikeline thickness will default to underline thickness, if set.
+
+=item *
+
+C<overline_thickness>, C<overline_position> - Overrides the font
+specified or calculated values for overline thickness and/or position.
+
+This may improve results for some fonts.
+
+Note that overline thickness will default to underline thickness, if
+set.
 
 =back
 
@@ -125,12 +184,11 @@ sub register_font {
     $props = pop(@_) if UNIVERSAL::isa( $_[-1], 'HASH' );
     my ( $font, $family, $style, $weight ) = @_;
 
-    if ( $style && !$weight && $style =~ s/^bold//i ) {
-	$weight = "bold";
+    if ( $style && !$weight && $style =~ s/^(heavy|bold|semi(?:bold)?|medium|book|light)//i ) {
+	$weight = $1;
     }
     $style  = _norm_style( $style   // "normal" );
     $weight = _norm_weight( $weight // "normal" );
-
     my $ff;
     if ( $font =~ /\.[ot]tf$/ ) {
 	if ( $font =~ m;^/; ) {
@@ -294,38 +352,44 @@ sub find_font {
     my $atts;
     $atts = pop(@_) if UNIVERSAL::isa( $_[-1], 'HASH' );
     my ( $family, $style, $weight ) = @_;
+    warn("find_font( $family, $style, $weight )\n") if $debug;
 
     my $try = sub {
-      if ( $fonts{$family}
-	 && $fonts{$family}->{$style}
-	 && $fonts{$family}->{$style}->{$weight} ) {
-	my $ff;
-	if ( $ff = $fonts{$family}->{$style}->{$weight}->{font} ) {
-	    return Text::Layout::FontDescriptor->new
-	      ( font   => $ff,
-		family => $family,
-		style  => $style,
-		weight => $weight,
-		shaping => $fonts{$family}->{$style}->{$weight}->{shaping},
-		interline => $fonts{$family}->{$style}->{$weight}->{interline},
-	      );
+	if ( $fonts{$family}
+	     && $fonts{$family}->{$style}
+	     && $fonts{$family}->{$style}->{$weight} ) {
+
+	    my $ff = $fonts{$family}->{$style}->{$weight};
+	    my %i = ( family => $family,
+		      style  => $style,
+		      weight => $weight );
+;
+	    if ( $ff->{font} ) {
+		$i{font} = $ff->{font};
+	    }
+	    elsif ( $ff->{loader_data} ) {
+		$i{loader_data} = $ff->{loader_data};
+		$i{loader}      = $loader;
+		$i{cache}       = $ff;
+	    }
+	    else {
+		return;
+	    }
+
+	    for ( qw( shaping ascender descender nosubset direction language
+		      underline_thickness underline_position
+		      strikeline_thickness strikeline_position
+		      overline_thickness overline_position
+		   ) ) {
+		$i{$_} = $ff->{$_};
+	    }
+
+	    if ( $debug ) {
+		warn("found( $i{family}, $i{style}, $i{weight} ) -> ",
+		     $i{loader_data}, "\n");
+	    }
+	    return Text::Layout::FontDescriptor->new(%i);
 	}
-	elsif ( $ff = $fonts{$family}->{$style}->{$weight}->{loader_data} ) {
-	    return Text::Layout::FontDescriptor->new
-	      ( loader_data => $ff,
-		loader => $loader,
-		cache  => $fonts{$family}->{$style}->{$weight},
-		family => $family,
-		style  => $style,
-		weight => $weight,
-		shaping => $fonts{$family}->{$style}->{$weight}->{shaping},
-		interline => $fonts{$family}->{$style}->{$weight}->{interline},
-	     );
-	}
-	else {
-	    return;
-	}
-      }
     };
 
     $style  = _norm_style( $style   // "normal" );
@@ -356,8 +420,8 @@ On Linux, fallback using fontconfig.
 
 =cut
 
-my $stylep  = qr/^( (?:bold)? (?:oblique|italic)  )$/ix;
-my $weightp = qr/^( (?:bold)  (?:oblique|italic)? )$/ix;
+my $stylep  = qr/^( (?:heavy|bold|semi(?:bold)?|medium|book|light)? (?:oblique|italic)  )$/ix;
+my $weightp = qr/^( (?:heavy|bold|semi(?:bold)?|medium|book|light)  (?:oblique|italic)? )$/ix;
 
 sub from_string {
     shift if UNIVERSAL::isa( $_[0], __PACKAGE__ );
@@ -405,7 +469,7 @@ sub parse {
 	    $style = "italic";
 	}
 	elsif ( $t =~ $weightp ) {
-	    $weight = "bold";
+	    $weight = $1;
 	}
 	elsif ( $t eq "normal" ) {
 	    $style = $weight = "";
@@ -447,12 +511,12 @@ sub from_filename {
 		 ( .*? )
 		 -?
 		 (roman?|normal|regular)?
-		 (light|book|bold)?
+		 (light|book|medium|semi(?:bold)?|bold|heavy)?
 		 (italic|ital|oblique|obli)?
 		 $/ix ) {
 	$family = $1       if $1;
 	$style  = "italic" if $4;
-	$weight = "bold"   if $3 && $3 =~ /^(bold)$/;
+	$weight = $3       if $3;
     }
 
     my $fd = Text::Layout::FontDescriptor->new
@@ -494,7 +558,7 @@ sub _norm_style {
 sub _norm_weight {
     my ( $weight ) = @_;
     $weight = lc $weight;
-    return "bold" if $weight =~ $weightp;
+    return $1 if $weight =~ $weightp;
 
     carp("Unhandled font weight: $weight\n")
       unless $weight =~ /^(regular|normal)?$/;
@@ -516,6 +580,7 @@ sub _fallback {
     return unless $fallback;
 
     my ( $family, $style, $weight ) = @_;
+    warn("fallback( $family, $style, $weight )\n") if $debug;
 
     my $pattern = $family;
     $pattern .= ":$style" if $style;
@@ -571,13 +636,20 @@ See L<Text::Layout>, L<Text::Layout::FontDescriptor>, L<HarfBuzz::Shaper>.
 sub _dump {
     foreach my $family ( sort keys %fonts ) {
 	foreach my $style ( qw( normal italic ) ) {
-	    foreach my $weight ( qw( normal bold ) ) {
+	    foreach my $weight ( qw( normal light book medium semi semibold bold heavy ) ) {
 		my $f = $fonts{$family}{$style}{$weight};
 		next unless $f;
 		printf STDERR ( "%-13s %s%s%s%s%s %s\n",
 				$family,
 				$style eq 'normal' ? "-" : "i",
-				$weight eq 'normal' ? "-" : "b",
+				$weight eq 'bold' ? "b"
+				: $weight eq 'light' ? "l"
+				: $weight eq 'book' ? "k"
+				: $weight eq 'medium' ? "m"
+				: $weight eq 'semi' ? "s"
+				: $weight eq 'semibold' ? "s"
+				: $weight eq 'heavy' ? "h"
+				: "-",
 				$f->{shaping} ? "s" : "-",
 				$f->{interline} ? "l" : "-",
 				$f->{font} ? "+" : " ",
