@@ -8,9 +8,9 @@ package Text::Layout::FontDescriptor;
 
 use Carp;
 
+use Text::Layout::Version;
 
-
-our $VERSION = "0.018.3";
+our $VERSION = $Text::Layout::VERSION;
 
 =head1 NAME
 
@@ -90,9 +90,7 @@ sub get_font {
     $self->{font} ||= do {
 	croak("Forgot to pass a layout context to get_font?")
 	  unless UNIVERSAL::isa( $context, 'Text::Layout' );
-	croak("Cannot load font $self (no loader defined)")
-	  unless $self->{loader};
-	$self->{loader}->( $self, $context, $self->{loader_data} );
+	$context->load_font( $self->{loader_data}, $self );
     };
 }
 
@@ -149,6 +147,26 @@ sub get_size {
     $self->{size};
 }
 
+=item set_direction
+
+=item get_direction
+
+Sets/gets the direction property of the font.
+
+=back
+
+=cut
+
+sub set_direction {
+    my ( $self, $direction ) = @_;
+    $self->{direction} = $direction;
+}
+
+sub get_direction {
+    my ( $self ) = @_;
+    $self->{direction};
+}
+
 # Not documented -- internal use only.
 
 sub set_shaping {
@@ -173,6 +191,31 @@ sub get_interline {
     $self->{interline};
 }
 
+# Not documented -- internal use only.
+
+# Note that the ascender/descender values are filled in at load time,
+# unless overridden by a set_... call.
+
+sub set_ascender {
+    my ( $self, $asc ) = @_;
+    $self->{ascender} = $asc;
+}
+
+sub get_ascender {
+    my ( $self ) = @_;
+    $self->{ascender};
+}
+
+sub set_descender {
+    my ( $self, $desc ) = @_;
+    $self->{descender} = $desc;
+}
+
+sub get_descender {
+    my ( $self ) = @_;
+    $self->{descender};
+}
+
 =over
 
 =item to_string
